/*
 * Servocontroller 1.9
 * © Fenna Pel, fenna@pereltje.nl
 * 
 * based on this article:
 * https://learn.adafruit.com/multi-tasking-the-arduino-part-2/overview by Bill Earl
 * 
 * The code is adapted and partialy rewritten to control upto 7 servo's with controlable speed and range.
 * Around the modifications new code is written to handle the specific properties of the used servo's and to add functionality.
 * 
 * additionaly there is 
 *   - a sketch to setup the internal EEPROM with sensible startup values
 *   - a hardware schematic with wiring details and general hardware setup
 * 
 * general advise: 
 *     This is NOT a newby project! 
 *     Before taking on this project, work yourself through the article by Bill Earl i mentioned earlier. 
 *     That will get you a basic idea of what is going on here.
 * 
 * 
 * changelog:
 * 
 * for version 1.9 (2020-11-24)
 *   - rearranged some code in the setup function to minimise 'noise' to the 3-8 decoder, 
 *     also in hardware rewired the E2 on the 3-8 to the +5v for the servo's. !E0 and !E1 are hard wired to ground. Check the 74hc238 datasheet for wiring details
 *   - Added code to display the currently used EEPROM block and write count value (only if the i2c lcd is connected)
 * 
 * for version 1.8 (2020-11-23)
 *   - removed serial monitor code as the i2c lcd gives enough debugging feedback
 *   - renamed the 'servostop' function to 'updateeeprom' to better discribe its true function
 *   - finished translating the commentary into english
 *   - removed version numbers from eeprom setting sketch
 * 
 * for version 1.7 (2020-11-22)
 *   - added reversedelay (controllable with potmeter8 when no programming leds are active)
 *   - added I2C lcd code for debugging, using the 'hd44780' library: https://www.arduino.cc/reference/en/libraries/hd44780/
 *   - added code to constrain the servo position within the set servo range. prevents hanging servo's while setting a smaller range
 *   - modified code to prevent 'illegal' values for position and range, keeps the servo's alive
 *   - added a seperate sketch for initialising the EEPROM (same version number as this sketch. It is basicaly a stripped down version of this sketch to just write some sensible starting values into the EEPROM
 *   
 *   
 * for version 1.6 (2020-11-21)
 *   - Added the EEPROM read and write code
 *   - modified the sweeper class to export the current position of the servo
 *   - modified the sweeper class to take a starting position
 *   - removed the flasher class as it is no longer needed to indicate a pending shutdown
 *   - added a serial print routine to debug the read EEPROM values 
 *   - starting to (re)write commentary in English as an international audience is starting to show interrest
 * 
 * for version 1.5 (2020-11-17)
 *   - further code clean-up
 *   - change 'int' types into 'short' where possible. preparing for storage into EEPROM and save some RAM usage
 *   - servo control with array's instead of regular variables
 *   - changed some hardcoded constants to system variables
 *   - first preparations for range programming (placeholder subroutines for reading andf writing the EEPROM)
 *   - code rewritten into the sketch for a serial monitor
 *   - added a function for monitoring a servo on the serial monitor 
 *   - At every timer interrupt only 1 servo is updated instead of all 7 at once (gives a smoother stroke on all servo's) A counter keeps track of the servo to update
 *   
 * for version 1.4
 *   - only changed the version name. In retrospect, can't remember exactly if i changed anything in the code
 * 
 * for version 1.3 (2020-10-31)
 *   - code clean-up and changed hardoded values for pins to predefined variables and constants
 *   - prepared the code for 7 servo's instead of the 2 hardcoded servo's
 *   - gracefull shutdown now gives a blinking power led and is a seperate function
 *   - removed code for serial monitor. An I2C lcd display can eventualy be added in the code
 *   
 * for version 1.2 and earlier
 *   - building up the first core of the code and structure of the sketch, gathering thoughts
 *   
 */



#include <Servo.h> 
#include <Wire.h>
#include <Adafruit_ADS1015.h>
#include <EEPROM.h>
#include <hd44780.h>                       // main hd44780 header
#include <hd44780ioClass/hd44780_I2Cexp.h> // i2c expander i/o class header

// system variables and constants
const short _SERVORANGE_MIN = 1000;
const short _SERVORANGE_MAX = 2000;
const short _SERVORANGE = _SERVORANGE_MAX - _SERVORANGE_MIN;
const short _SERVOCENTRE = (_SERVORANGE_MIN + _SERVORANGE_MAX)/2;
short _BOARDPOWER; // also used in the servo class
const short _START_DELAY = 1000;
const short _INCREMENT = 3;
short _REVERSE_DELAY = 1000;
short _REVERSE_DELAY_MIN = 1000;
short _REVERSE_DELAY_MAX = 9999;
int _DELAY_FLAG[7]= {LOW,LOW,LOW,LOW,LOW,LOW,LOW};
short _INTERRUPT_COUNTER = 0;
const short _ADA_ADC_5V = 1680; // the 5 volt value for the adafruit I2C ADC
const short _ARD_ADC_5V = 1023; // the 5 volt value for arduino's onboard ADC, for code consistancy
const short _SWEEP_MIN = 1;
const short _SWEEP_MAX = 50;
const short _ADS1015_I2C_1 = 0x48; // I2C address of the used adafruit I2C ADC, for wiring details: https://learn.adafruit.com/adafruit-4-channel-adc-breakouts/
const short _KEY_DELAY=500;
bool _I2C_LCD_PRESENT = true;

// pin assignments
short potmeterpin[7] = {A0,A1,A2,A3,0,1,2};
short servopin[7] = {6,7,8,9,10,11,12};
short potmeterpin8 = 3;
short ledpin = 13; // powerled
short ledadr0 = 3; // these 3 pins are used by the 3-to-8 decoder for the programming leds
short ledadr1 = 4; //
short ledadr2 = 5; //
short poweroffpin = A7;
short programpin = A6;
short servoonpin = 2;

// EEPROM handling:
unsigned short _EEPROM_POSITION[7] = {_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE,_SERVOCENTRE};
unsigned int _EEPROM_WRITE_COUNT; // Keep track of the number of write cycles to the EEPROM
byte _EEPROM_BLOCK_NUMBER; // Keep track of which EEPROM block is used
int _EEPROM_BLOCK_START_ADDRESS;
const int _EEPROM_MAX_WRITES = 90000;

// variables with sensible startup values:
short potmeter[7] = {0,0,0,0,0,0,0};
short potmeter8 = 0;
short timeoff_Pon = 1;
short timeoff_Poff = 100;
short timeon = 100;
short ledadr = 7;
short programkeyvalue = 0;
short programkey = 0;
short programkey2 = 0;
short servospeed[7] = {_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN};
short servorange[8] = {_SERVORANGE,_SERVORANGE,_SERVORANGE,_SERVORANGE,_SERVORANGE,_SERVORANGE,_SERVORANGE,_REVERSE_DELAY_MIN};

const int LCD_COLS = 20;
const int LCD_ROWS = 4;

// -------------------------------------------------------------
class Sweeper
{
  Servo servo;              // the servo
  int  updateInterval;      // interval between updates
  
  volatile short pos;                  // current servo position 
  volatile short pos_exec ;
  volatile unsigned long lastUpdate; // last update of position
  volatile short increment;            // increment to move for each interval
  int sweeperdelayflag;

public: 
  Sweeper(short interval)
  {
    updateInterval = interval;
    increment = _INCREMENT;
  }
  
  void Attach(short pin)
  {
    servo.attach(pin);
  }
  
  void Detach()
  {
    servo.detach();
  }
  
  void refresh(short newinterval, short updateservorange)
  {
    if (sweeperdelayflag == LOW) updateInterval=newinterval;
  }
  
  void Start (short startposition)
  {
     pos = startposition;
  }
  
  //void Update(unsigned long currentMillis, int updateservorange, int updatedelayflag)
  void Update(unsigned long currentMillis, int updateservorange)
  {
    int posmax = _SERVOCENTRE+(updateservorange/2);
    int posmin = _SERVOCENTRE-(updateservorange/2);
    pos = constrain(pos,posmin,posmax);
    if((currentMillis - lastUpdate) > updateInterval)  // time to update
    {
      sweeperdelayflag = LOW;
      if (_BOARDPOWER == LOW)
      {
        updateservorange = 0;
      }
      lastUpdate = currentMillis;
      pos += increment;
      servo.writeMicroseconds(pos);
      if ((pos >= posmax) || (pos <= posmin)) // end of sweep
      {
        increment = -increment;
        updateInterval = _REVERSE_DELAY;
        sweeperdelayflag = HIGH;
      }
    }
    
  }
  short currentposition() {return pos;} // to export the current position to outside the class
};

// ------------------------------------------
Adafruit_ADS1015 ads1015(_ADS1015_I2C_1); 
Sweeper sweeper[7](0);
hd44780_I2Cexp lcd; // declare lcd object: auto locate & auto config expander chip

// ------------------------------------------------------
// the nitty gritty of updating the EEPROM
void updateeeprom()
{
  //increment write counter of the current block, max out at _EEPROM_MAX_WRITES and store at beginning of the current block
  _EEPROM_WRITE_COUNT = EEPROM.read(_EEPROM_BLOCK_START_ADDRESS+1)*65536 + EEPROM.read(_EEPROM_BLOCK_START_ADDRESS+2)*256 + EEPROM.read(_EEPROM_BLOCK_START_ADDRESS+3);
  _EEPROM_WRITE_COUNT++;
  if ( _EEPROM_WRITE_COUNT > _EEPROM_MAX_WRITES) // prepare for a new block in EEPROM
  {
    _EEPROM_BLOCK_NUMBER++;
    _EEPROM_WRITE_COUNT = 0;
    _EEPROM_BLOCK_START_ADDRESS = _EEPROM_BLOCK_NUMBER * 32;
  }
  EEPROM.update(0, _EEPROM_BLOCK_NUMBER);
  EEPROM.update(_EEPROM_BLOCK_START_ADDRESS+1, _EEPROM_WRITE_COUNT/65536);
  EEPROM.update(_EEPROM_BLOCK_START_ADDRESS+2, (_EEPROM_WRITE_COUNT%65536)/256);
  EEPROM.update(_EEPROM_BLOCK_START_ADDRESS+3, _EEPROM_WRITE_COUNT%256);

  // store servo values:
  for (byte i = 0; i <= 6; i++) 
  {
    _EEPROM_POSITION[i]=sweeper[i].currentposition();
    int base_address = ((1+i)*4+_EEPROM_BLOCK_START_ADDRESS);
    EEPROM.update(base_address, _EEPROM_POSITION[i]/256);
    EEPROM.update(base_address+1, _EEPROM_POSITION[i]%256);
    EEPROM.update(base_address+2, servorange[i]/256);
    EEPROM.update(base_address+3, servorange[i]%256);
  }
}

// -----------------------------------------------------
// shutdown control function
void gracefullshutdown()
{
  digitalWrite(ledadr0, HIGH);
  digitalWrite(ledadr1, HIGH);
  digitalWrite(ledadr2, HIGH);
  updateeeprom();
  if (_I2C_LCD_PRESENT)
  {
    lcd.clear();
    lcd.setCursor (0,0);
    lcd.print(_EEPROM_BLOCK_NUMBER);
    lcd.setCursor (0,1);
    lcd.print (_EEPROM_WRITE_COUNT);
  }
  delay(_START_DELAY*2);
  digitalWrite (ledpin, LOW);
  digitalWrite(servoonpin, LOW);
}

// --------------------------------------------
// display the debug values on a 20x4 i2c lcd
void printservos()
{
  lcd.setCursor(0,0);
  lcd.print (servorange[0]);
  lcd.print ("   ");
  lcd.setCursor(0,1);
  lcd.print (sweeper[0].currentposition());
  lcd.print ("   ");
  lcd.setCursor(5,0);
  lcd.print (servorange[1]);
  lcd.print ("   ");
  lcd.setCursor(5,1);
  lcd.print (sweeper[1].currentposition());
  lcd.print ("   ");
  lcd.setCursor(10,0);
  lcd.print (servorange[2]);
  lcd.print ("   ");
  lcd.setCursor(10,1);
  lcd.print (sweeper[2].currentposition());
  lcd.print ("   ");
  lcd.setCursor(15,0);
  lcd.print (servorange[3]);
  lcd.print ("   ");
  lcd.setCursor(15,1);
  lcd.print (sweeper[3].currentposition());
  lcd.print ("   ");
  lcd.setCursor(0,2);
  lcd.print (servorange[4]);
  lcd.print ("   ");
  lcd.setCursor(0,3);
  lcd.print (sweeper[4].currentposition());
  lcd.print ("   ");
  lcd.setCursor(5,2);
  lcd.print (servorange[5]);
  lcd.print ("   ");
  lcd.setCursor(5,3);
  lcd.print (sweeper[5].currentposition());
  lcd.print ("   ");
  lcd.setCursor(10,2);
  lcd.print (servorange[6]);
  lcd.print ("   ");
  lcd.setCursor(10,3);
  lcd.print (sweeper[6].currentposition());
  lcd.print ("   ");
  lcd.setCursor(15,2);
  lcd.print (potmeter8);
  lcd.print ("  ");
  lcd.setCursor(15,3);
  lcd.print (_REVERSE_DELAY);
  lcd.print (" ");
}

// --------------------------------------------
void setup() 
{
  pinMode(ledadr0, OUTPUT);
  pinMode(ledadr1, OUTPUT);
  pinMode(ledadr2, OUTPUT);
  pinMode(ledpin, OUTPUT); 
  pinMode(servoonpin, OUTPUT);
  pinMode(poweroffpin, INPUT);

  digitalWrite(ledadr0, HIGH);
  digitalWrite(ledadr1, HIGH);
  digitalWrite(ledadr2, HIGH);
  
  int status;
  status = lcd.begin(LCD_COLS, LCD_ROWS);
  if(status) // non zero status means it was unsuccesful
  {
    _I2C_LCD_PRESENT = false;
  }
  if (_I2C_LCD_PRESENT) lcd.print("Servo controller 1.9");
  
  _EEPROM_BLOCK_NUMBER = EEPROM.read(0); // determine which block is used and calculate block offset (the current blocknumber is stored at the first EEPROM address)
  _EEPROM_BLOCK_START_ADDRESS = _EEPROM_BLOCK_NUMBER * 32; 
  
  // get most recent values from EEPROM
  for (byte i = 0; i <= 6; i++) 
  {
    int base_address = ((1+i)*4+_EEPROM_BLOCK_START_ADDRESS);
    _EEPROM_POSITION[i] = EEPROM.read(base_address) * 256 + EEPROM.read(base_address+1);
    servorange[i] =  EEPROM.read(base_address+2) * 256 + EEPROM.read(base_address+3);
    sweeper[i].Attach(servopin[i]);
    sweeper[i].Start(_EEPROM_POSITION[i]);
  }

  ads1015.begin();
  
  // Timer0 is already used for millis() - we'll just interrupt somewhere
  // in the middle and call the "Compare A" function below
  OCR0A = 0xAF;
  TIMSK0 |= _BV(OCIE0A);
  
  //gracefull start
  delay(_START_DELAY);
  digitalWrite(servoonpin, HIGH); // switching the power to the servo's on
  digitalWrite (ledpin, HIGH); // activating the power indicating led
} 

// Interrupt is called once a millisecond, The servo's are updated 1 per interrupt call. A counter keeps track of which servo is to be updated
SIGNAL(TIMER0_COMPA_vect) 
{
  unsigned long currentMillis = millis();
  sweeper[_INTERRUPT_COUNTER].Update(currentMillis,servorange[_INTERRUPT_COUNTER]);
  _INTERRUPT_COUNTER++;
  if (_INTERRUPT_COUNTER == 7) _INTERRUPT_COUNTER = 0;
} 

// -------------------------------------------
void loop()
{
  for (short i = 0; i <= 3; i++) 
  {
    potmeter[i] = analogRead(potmeterpin[i]);
    servospeed[i] = map(potmeter[i], 0,_ARD_ADC_5V,_SWEEP_MIN,_SWEEP_MAX);
  }

  for (short i = 4; i <= 6; i++) 
  {
    potmeter[i] = ads1015.readADC_SingleEnded(potmeterpin[i]);
    servospeed[i] = map(potmeter[i], 0,_ADA_ADC_5V,_SWEEP_MIN,_SWEEP_MAX);
  }

  if (_I2C_LCD_PRESENT) printservos();

  for (short i = 0; i <= 6; i++) 
  {
    sweeper[i].refresh(servospeed[i],servorange[i]);
  }

  _BOARDPOWER = analogRead (poweroffpin);
  if (_BOARDPOWER == LOW) gracefullshutdown();

  programkeyvalue = analogRead(programpin);
  programkey = map(programkeyvalue, 1,1024,0,3);
  
  digitalWrite(ledadr0, bitRead(ledadr,0));
  digitalWrite(ledadr1, bitRead(ledadr,1));
  digitalWrite(ledadr2, bitRead(ledadr,2));
  delay(_KEY_DELAY);
  programkeyvalue = analogRead(programpin);
  programkey2 = map(programkeyvalue, 1,1024,0,3);
  if (programkey == 2 && programkey2 == 2) ledadr = ledadr+1;
  if (programkey == 1 && programkey2 == 1)
  {
    ledadr = 7;
    updateeeprom();
  }
  if (ledadr == 8) ledadr=0;
  potmeter8 = ads1015.readADC_SingleEnded(potmeterpin8);
  if (ledadr != 7) servorange[ledadr] = map(potmeter8, 0,_ADA_ADC_5V,(_SERVORANGE/10),_SERVORANGE);
  if (servorange[ledadr] >= _SERVORANGE) servorange[ledadr] = _SERVORANGE;
  if (ledadr == 7) _REVERSE_DELAY = map(potmeter8, 0,_ADA_ADC_5V,_REVERSE_DELAY_MIN,_REVERSE_DELAY_MAX);  
  _REVERSE_DELAY = constrain(_REVERSE_DELAY,_REVERSE_DELAY_MIN,_REVERSE_DELAY_MAX);
}

// ---------------------------------------------------------
